-- Conversation Manager

--=== Local junk ===========================================================--

local function cMessage(msg)
	if (DEFAULT_CHAT_FRAME) then
		DEFAULT_CHAT_FRAME:AddMessage(msg);
	end
end

local function cColourString(r, g, b, bold)

	if bold then
		r = (r + 1) / 2;
		g = (g + 1) / 2;
		b = (b + 1) / 2;
	end

	return string.format("|cff%02x%02x%02x", r * 255, g * 255, b * 255);
end

local function cChatColourString(chattype, bold)
	if not ChatTypeInfo[chattype] then
		return "";
	end

	return cColourString(ChatTypeInfo[chattype].r, ChatTypeInfo[chattype].g, ChatTypeInfo[chattype].b, bold);
end

--===--

local cText = "";		-- This will contain whatever is currently in the window
local cNumItems = 0;		-- Number of 'items' on the side list
local cSelected = 1;		-- Currently selected
local cItems = {};		-- Table of sidebar items
local cChannels = {};		-- Table of numbered channels we are currently in. Stored in lowercase
local cChannelNames = {};	-- Table of players in each channel
local cServerChannels = {};	-- Table of server channels
local cLines = 0;		-- Number of lines in currently viewed history
local cBufferLines = 0;		-- Number of lines in buffer of current channel

local cChannelKludge = "";	-- Since CHAT_MSG_CHANNEL_NOTICE seems to fire BEFORE it actually removes you from the channel

local cFinishedLoading = nil;

local cDateStamp = "";		-- Constant datestamp generated per login

local cPaused = nil;		-- Paused?
local cPasting = nil;
local cPasteCtype = "";
local cPasteCtarget = "";

local RealmName = "";
local PlayerName = "";
local GuildName = "";

cIsWOTLK = nil;

--=== Initialization =======================================================--

function Converse_OnLoad()
	-- So ESC can close it...
	tinsert(UISpecialFrames, "ConverseFrame");

	this:RegisterForDrag("LeftButton");

	Converse_InitHooks();

	-- WotLK check
	local ver, build = GetBuildInfo();
	if string.sub(ver,1,1) == "3" then
		cIsWOTLK=true;
		DEFAULT_CHAT_FRAME:AddMessage("Conv: WOTLK detected");
	end
end

function Converse_InitHooks()
	this:RegisterEvent("VARIABLES_LOADED");
	this:RegisterEvent("GUILD_ROSTER_UPDATE");
	this:RegisterEvent("FRIENDLIST_UPDATE");

	-- Chat events
	this:RegisterEvent("CHAT_MSG_WHISPER");
	this:RegisterEvent("CHAT_MSG_WHISPER_INFORM");

	-- RealID fun!
	this:RegisterEvent("CHAT_MSG_BN_WHISPER");
	this:RegisterEvent("CHAT_MSG_BN_WHISPER_INFORM");

	this:RegisterEvent("CHAT_MSG_AFK");
	this:RegisterEvent("CHAT_MSG_DND");
	this:RegisterEvent("CHAT_MSG_GUILD");
	this:RegisterEvent("CHAT_MSG_OFFICER");

	this:RegisterEvent("GUILD_MOTD");

	this:RegisterEvent("CHAT_MSG_PARTY");
	this:RegisterEvent("CHAT_MSG_PARTY_LEADER");

	this:RegisterEvent("CHAT_MSG_RAID");
	this:RegisterEvent("CHAT_MSG_RAID_LEADER");
	this:RegisterEvent("CHAT_MSG_RAID_WARNING");

	this:RegisterEvent("CHAT_MSG_BATTLEGROUND");
	this:RegisterEvent("CHAT_MSG_BATTLEGROUND_LEADER");
	this:RegisterEvent("CHAT_MSG_BG_SYSTEM_ALLIANCE");
	this:RegisterEvent("CHAT_MSG_BG_SYSTEM_HORDE");
	this:RegisterEvent("CHAT_MSG_BG_SYSTEM_NEUTRAL");

	this:RegisterEvent("CHAT_MSG_SAY");
	this:RegisterEvent("CHAT_MSG_EMOTE");
	this:RegisterEvent("CHAT_MSG_TEXT_EMOTE");
	this:RegisterEvent("CHAT_MSG_YELL");

	this:RegisterEvent("CHAT_MSG_MONSTER_SAY");
	this:RegisterEvent("CHAT_MSG_MONSTER_EMOTE");
	this:RegisterEvent("CHAT_MSG_MONSTER_YELL");

	this:RegisterEvent("CHAT_MSG_CHANNEL");
	this:RegisterEvent("CHAT_MSG_CHANNEL_JOIN");
	this:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE");
	this:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");
	this:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER");

	-- This could be messy...
	this:RegisterEvent("CHAT_MSG_CHANNEL_LIST");

	SlashCmdList["CONVERSE"] =
		function(args)
			Converse_SlashCmd(args)
		end
	SlashCmdList["CHATALL"] =
		function(args)
			Converse_RefreshChannelPlayers();
		end

end

function Converse_InitData()
	-- Setup local constants
	RealmName = GetCVar("realmname");
	PlayerName = UnitName("player");

	-- Wheee....
	if (IsInGuild()) then
		GuildRoster();
	end
	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end

	-- *** Options ***
	if not ConverseState		then ConverseState = {}; end
	if not ConverseState[RealmName]	then ConverseState[RealmName] = {}; end

	if not ConverseState[RealmName][PlayerName] then
		ConverseState[RealmName][PlayerName] = {};
		-- A few default settings
		ConverseState[RealmName][PlayerName].hidden = 1;
		ConverseState[RealmName][PlayerName].chanblink = 1;
		ConverseState[RealmName][PlayerName].msgblink = 1;
		ConverseState[RealmName][PlayerName].syncchat = 1;
		ConverseState[RealmName][PlayerName].datestamp = 1;
		ConverseState[RealmName][PlayerName].gemfilter = nil;
	end

	if not ConverseState[RealmName][PlayerName].fontsize then
		ConverseState[RealmName][PlayerName].fontsize = CONVERSE_DEF_FONTSIZE;
	end
	if not ConverseState[RealmName][PlayerName].lastselected then
		ConverseState[RealmName][PlayerName].lastselected = CONVERSE_DEF_SELECTED;
	end
	--if not ConverseState[RealmName][PlayerName].historylock then
	--	ConverseState[RealmName][PlayerName].historylock = 1;
	--end
	if not ConverseState[RealmName][PlayerName].maxlines then
		ConverseState[RealmName][PlayerName].maxlines = CONVERSE_DEF_MAXLINES;
	end
	--if not ConverseState[RealmName][PlayerName].blink then
	--	ConverseState[RealmName][PlayerName].blink = 1;
	--end
	--if not ConverseState[RealmName][PlayerName].chanblink then
	--	ConverseState[RealmName][PlayerName].chanblink = 1;
	--end
	--if not ConverseState[RealmName][PlayerName].msgblink then
	--	ConverseState[RealmName][PlayerName].msgblink = 1;
	--end
	if not ConverseState[RealmName][PlayerName].ignore then
		ConverseState[RealmName][PlayerName].ignore = {};
		ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] = 1;
	end

	-- TODO: Maybe we could set up an 'alts' list, so that if someone
	-- switches characters you can still carry on a seamless conversation

	-- *** ConverseHistory ***
	if not ConverseHistory then ConverseHistory = {}; end
	if not ConverseHistory[RealmName] then ConverseHistory[RealmName] = {}; end

	-- Setup the default channels
	-- Say
	if not ConverseHistory[RealmName][CONVERSE_INDEX_SAY] then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_SAY].buffer then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY].buffer = {};
	end
	-- Party
	if not ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] then
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].buffer then
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].buffer = {};
	end
	-- Raid
	if not ConverseHistory[RealmName][CONVERSE_INDEX_RAID] then
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_RAID].buffer then
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID].buffer = {};
	end
	-- BG
	if not ConverseHistory[RealmName][CONVERSE_INDEX_BG] then
		ConverseHistory[RealmName][CONVERSE_INDEX_BG] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_BG].text then
		ConverseHistory[RealmName][CONVERSE_INDEX_BG].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_INDEX_BG].buffer then
		ConverseHistory[RealmName][CONVERSE_INDEX_BG].buffer = {};
	end
	-- Guild
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].buffer then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].buffer = {};
	end
	-- Officer
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {}; 
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].buffer then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].buffer = {}; 
	end

	-- 10 channels
	Converse_RefreshChannelPlayers();

	-- Converse_RefreshChannelsList();
	--
	--local cname,value;
	--for cname,value in pairs(cChannels) do
	--	if not cServerChannels[cname] then
	--		ListChannelByName(cname);
	--	end
	--end

	if not ConverseState[RealmName][PlayerName].multipaste then
		ConverseState[RealmName][PlayerName].multipaste = "";
	end
end

function Converse_InitState()
	local i;

	--cDateStamp = CONVERSE_DATESTAMP_PRE..date()..CONVERSE_DATESTAMP_POST;
	cDateStamp = string.format(CONVERSE_DATESTAMP_FORMAT, PlayerName, date());

	ConverseItems_Rebuild();

	if (cSelected > cNumItems) then
		-- We have a discrepancy!
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
	end
	ConverseItems_Update();

	if (cNumItems > 0) and (cSelected > 0) then
		ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
		ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);

		-- And set up the history window
		ConverseHistory_GenerateText(ConverseState[RealmName][PlayerName].lastselected);
		ConverseHistory_Update();
	else
		-- BLANK!
		ConverseFrameItemsSelectedText:SetText("");
		cText = "";
		cLines = 0;
		cBufferLines = 0;
		ConverseHistory_Update();
	end


	-- Font size
	ConverseHistory_SetFontSize(ConverseState[RealmName][PlayerName].fontsize);

	-- Paste frame text
	ConverseFrameMultipasteFrameScrollFrameText:SetText(ConverseState[RealmName][PlayerName].multipaste);

	-- 'Finished' loading.
	cFinishedLoading = 1;
	DEFAULT_CHAT_FRAME:AddMessage(CONVERSE_MSG_LOADED);

	-- Set up text labels and all that fun stuff
	ConverseFrameTitleText:SetText(CONVERSE_HEADER_TITLE.." ("..CONVERSE_VER..") - "..RealmName);
	ConverseFrameIgnoreText:SetText(CONVERSE_HEADER_IGNORE);

	-- Ignore checkbuttons
	for i=1,CONVERSE_MAXCHANNELS do
		getglobal("ConverseFrameIgnoreCheckButton"..i.."Text"):SetText(i);
		if ConverseState[RealmName][PlayerName].ignore[i] then
			getglobal("ConverseFrameIgnoreCheckButton"..i):SetChecked(1);
		end
	end
	-- Ignore say
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY.."Text"):SetText(CONVERSE_BUTTON_IGNORESAY);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY):SetChecked(1);
	end
	-- Ignore party
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY.."Text"):SetText(CONVERSE_BUTTON_IGNOREPARTY);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY):SetChecked(1);
	end
	-- Ignore raid
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID.."Text"):SetText(CONVERSE_BUTTON_IGNORERAID);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID):SetChecked(1);
	end
	-- Ignore BG
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_BG.."Text"):SetText(CONVERSE_BUTTON_IGNOREBG);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_BG):SetChecked(1);
	end
	-- Ignore guild
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD.."Text"):SetText(CONVERSE_BUTTON_IGNOREGUILD);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD):SetChecked(1);
	end
	-- Ignore officer
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER.."Text"):SetText(CONVERSE_BUTTON_IGNOREOFFICER);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER):SetChecked(1);
	end
	-- Ignore whispers
	getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER.."Text"):SetText(CONVERSE_BUTTON_IGNOREWHISPER);
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER):SetChecked(1);
	end

	-- Hook ChatFrameEditBox
	hooksecurefunc("ChatFrame_OpenChat", Converse_ChatFrame_OpenChat);

	-- Alt Editbox
	if ConverseState[RealmName][PlayerName].altedit then
		ConverseFrameAltEditFrame:Show();
	else
		ConverseFrameAltEditFrame:Hide();
	end

	-- Links panel
	if ConverseState[RealmName][PlayerName].links then
		ConverseFrameLinksPanel:Show();
	end
	ConverseFrameLinksPanelMessageFrame:SetFading(false);

	-- Window lock toggle
	if ConverseState[RealmName][PlayerName].locked then
		ConverseFrameLockButton:SetChecked(1);
	else
		ConverseFrameLockButton:SetChecked(0);
	end

	-- And is the frame visible or not
	if ConverseState[RealmName][PlayerName].hidden then
		ConverseFrame:Hide();
	else
		ConverseFrame:Show();
	end
end

function Converse_RefreshChannelsList()
	local i, cnum, cname, nname;
	local key, value;
	local button;
	local schan = {};

	-- Figure out the server channels
	cServerChannels = {};
	schan[1], schan[2], schan[3], schan[4], schan[5], schan[6], schan[7], schan[8], schan[9], schan[10] = EnumerateServerChannels();
	for i = 1,table.getn(schan) do
		-- Channel name starts with letter, but could possible contain digits...
		cname = string.lower(string.sub(schan[i], string.find(schan[i], "%a[%a%d]+")));
		cServerChannels[cname] = GetChannelName(cname);
	end

	-- Make our channels list
	cChannels = {};
	for i=1,CONVERSE_MAXCHANNELS do
		cnum, cname = GetChannelName(i);
		button = getglobal("ConverseFrameIgnoreCheckButton"..i);
		if cnum and (cnum ~= 0) and (cnum == i) then
			-- Remove everything after the first space (eg. " - Darnassus")
			nname = string.lower(string.sub(cname, string.find(cname,"[^ ]*")));
			-- Ignore the entry if it's the same as cChannelKludge
			if (nname ~= cChannelKludge) then
				cChannels[nname] = cnum;
			else
				-- Remove the history entry
				-- ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..nname] = nil;
			end
			-- Enable the button
			button:Enable();
			button:Show();
		else
			-- Disable the button
			button:SetChecked(0);
			button:Disable();
			button:Hide();
		end
	end

	-- Make sure history entries are there.
	for key, value in pairs(cChannels) do
		key = string.lower(key);
		if not ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key] then
			ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key] = {}
		end
		if not ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].text then
			ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].text = {};
		end
		if not ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].buffer then
			ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..key].buffer = {};
		end
	end
end

function Converse_RefreshChannelPlayers()
	Converse_RefreshChannelsList();

	local cname,value;
	for cname,value in pairs(cChannels) do
		if not cServerChannels[cname] then
			ListChannelByName(cname);
		end
	end
end

--=== Console interface ====================================================--
function Converse_SlashCmd(args)
	local i;

	if (args) then
		local opts = string.lower(args);
		if (opts == "") then
			Converse_ShowHelp();
		elseif (opts == CONVERSE_SLASH_HIDE) then
			ConverseFrame:Hide();
			ConverseState[RealmName][PlayerName].hidden = 1;
			PlaySound("igQuestListClose");
		elseif (opts == CONVERSE_SLASH_SHOW) then
			ConverseFrame:Show();
			ConverseState[RealmName][PlayerName].hidden = nil;
			PlaySound("igQuestListOpen");
		elseif (opts == CONVERSE_SLASH_LOCK) then
			ConverseState[RealmName][PlayerName].locked = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_LOCKED);
			ConverseFrameLockButton:SetChecked(1);
		elseif (opts == CONVERSE_SLASH_UNLOCK) then
			ConverseState[RealmName][PlayerName].locked = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_UNLOCKED);
			ConverseFrameLockButton:SetChecked(0);
		elseif (opts == CONVERSE_SLASH_RESET) then
			ConverseFrame:ClearAllPoints();
			ConverseFrame:SetPoint("TOPRIGHT", "UIParent", "TOPRIGHT", -16, -104);
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_RESET);
		elseif (string.find(opts, CONVERSE_SLASH_FONTSIZE.." ") == 1) then
			i = tonumber(string.sub(opts, string.len(CONVERSE_SLASH_FONTSIZE) + 2, string.len(opts)));
			if (i >= CONVERSE_FONTMIN) and (i <= CONVERSE_FONTMAX) then
				ConverseHistory_SetFontSize(i);
				ConverseState[RealmName][PlayerName].fontsize = i;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_FONTSIZE..i);
			else
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_FONTSIZE..i);
			end
		elseif (string.find(opts, CONVERSE_SLASH_MAXLINES.." ") == 1) then
			i = tonumber(string.sub(opts, string.len(CONVERSE_SLASH_MAXLINES) + 2, string.len(opts)));
			if (i == 0) then
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_MAXLINES..i);
				--ConverseState[RealmName][PlayerName].maxlines = 0;
				-- Infinite. BEWARE!
			elseif (i >= CONVERSE_LINESMIN) and (i <= CONVERSE_LINESMAX) then
				ConverseState[RealmName][PlayerName].maxlines = i;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_MAXLINES..i);
			else
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_ERR_MAXLINES..i);
			end
		-- Blinkinlights!
		elseif (opts == CONVERSE_SLASH_BLINK) then
			if not ConverseState[RealmName][PlayerName].blink then
				ConverseState[RealmName][PlayerName].blink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].blink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		elseif (opts == CONVERSE_SLASH_MSGBLINK) then
			if not ConverseState[RealmName][PlayerName].msgblink then
				ConverseState[RealmName][PlayerName].msgblink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].msgblink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		elseif (opts == CONVERSE_SLASH_CHANBLINK) then
			if not ConverseState[RealmName][PlayerName].chanblink then
				ConverseState[RealmName][PlayerName].chanblink = 1;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_ENABLED);
			else
				ConverseState[RealmName][PlayerName].chanblink = nil;
				cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_DISABLED);
			end
			ConverseItems_Rebuild();
			ConverseItems_Update();
		-- Clear
		elseif (opts == CONVERSE_SLASH_CLEARALL) then
			Converse_Clear(CONVERSE_CLEARALL);
		elseif (opts == CONVERSE_SLASH_CLEARCHAT) then
			Converse_Clear(CONVERSE_CLEARCHAT);
		elseif (opts == CONVERSE_SLASH_CLEARCHAN) then
			Converse_Clear(CONVERSE_CLEARCHAN);
		elseif (opts == CONVERSE_SLASH_CLEARMSG) then
			Converse_Clear(CONVERSE_CLEARMSG);
		elseif (opts == CONVERSE_SLASH_CLEARRAND) then
			Converse_Clear(CONVERSE_CLEARRAND);
		--elseif (opts == CONVERSE_SLASH_CHANREFRESH) then
		--	Converse_RefreshChannelPlayers();
		-- Chat sync
		elseif (opts == CONVERSE_SLASH_SYNC) then
			ConverseState[RealmName][PlayerName].syncchat = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_SYNC..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NOSYNC) then
			ConverseState[RealmName][PlayerName].syncchat = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_SYNC..CONVERSE_MSG_DISABLED);
		-- Date stamping
		elseif (opts == CONVERSE_SLASH_DATE) then
			ConverseState[RealmName][PlayerName].datestamp = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_DATE..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NODATE) then
			ConverseState[RealmName][PlayerName].datestamp = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_DATE..CONVERSE_MSG_DISABLED);
		-- GEM filter
		elseif (opts == CONVERSE_SLASH_GEMFILTER) then
			ConverseState[RealmName][PlayerName].gemfilter = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NOGEMFILTER) then
			ConverseState[RealmName][PlayerName].gemfilter = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_DISABLED);
		-- Alternate EditBox
		elseif (opts == CONVERSE_SLASH_ALTEDIT) then
			ConverseFrameAltEditFrame:Show();
			ConverseState[RealmName][PlayerName].altedit = 1;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_ALTEDIT..CONVERSE_MSG_ENABLED);
		elseif (opts == CONVERSE_SLASH_NOALTEDIT) then
			ConverseFrameAltEditFrame:Hide();
			ConverseState[RealmName][PlayerName].altedit = nil;
			cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_ALTEDIT..CONVERSE_MSG_DISABLED);
		-- EOL
		elseif (opts == CONVERSE_SLASH_STATUS) then
			Converse_Status();
		else
			Converse_ShowHelp();
		end
	else
		Converse_ShowHelp();
	end
end

function Converse_Status()
	cMessage(CONVERSE_MSG_STATUSHEADER);
	if ConverseState[RealmName][PlayerName].locked then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_LOCKED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_UNLOCKED);
	end
	cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_FONTSIZE..ConverseState[RealmName][PlayerName].fontsize);
	cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_MAXLINES..ConverseState[RealmName][PlayerName].maxlines);
	if ConverseState[RealmName][PlayerName].blink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_BLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].chanblink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_CHANBLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].msgblink then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_BLINKTYPE..CONVERSE_MSG_MSGBLINK..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].syncchat then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_SYNC..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_SYNC..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].datestamp then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_DATE..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_DATE..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].altedit then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_ALTEDIT..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_ALTEDIT..CONVERSE_MSG_DISABLED);
	end
	if ConverseState[RealmName][PlayerName].gemfilter then
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_ENABLED);
	else
		cMessage(CONVERSE_COLOUR_PRINT.."  "..CONVERSE_MSG_GEMFILTER..CONVERSE_MSG_DISABLED);
	end
end

function Converse_ShowHelp()
	local i;
		for i=1, table.getn(CONVERSE_HELPTEXT) do
		cMessage(CONVERSE_HELPTEXT[i]);
	end
end

function Converse_Clear(cleartype)
	if (cleartype == CONVERSE_CLEARALL) then
		ConverseHistory[RealmName] = {}
		Converse_InitData();
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARALL..RealmName);
		ConverseItems_Rebuild();
		ConverseItems_Update();
		cText = "";
		cLines = 0;
		ConverseHistory_Update();
		return;
	elseif (cleartype == CONVERSE_CLEARCHAT) then
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_SAY].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_RAID].text = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_BG] = {};
		ConverseHistory[RealmName][CONVERSE_INDEX_BG].text = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {};
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARCHAT..RealmName);
	elseif (cleartype == CONVERSE_CLEARCHAN) then
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			lname = string.lower(cname);
			if (pfx == CONVERSE_PREFIX_CHANNEL) then
				if cChannels[lname] then
					-- Numbered channel
					ConverseHistory[RealmName][key] = {};
					ConverseHistory[RealmName][key].text = {};
				else
					-- Unnumbered channel
					ConverseHistory[RealmName][key] = nil;
				end
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARCHAN..RealmName);
	elseif (cleartype == CONVERSE_CLEARMSG) then
		-- Messages only
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			if (pfx == CONVERSE_PREFIX_MSG) then
				ConverseHistory[RealmName][key] = nil;
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARMSG..RealmName);
	elseif (cleartype == CONVERSE_CLEARRAND) then
		local lfriends = {};
		for i=1,GetNumFriends() do
			local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
			if fname then
				if fconn then
					lfriends[fname] = 1;
				else
					lfriends[fname] = 0;
				end
			end
		end
		-- Non friend messages only
		local key, value;
		for key, value in pairs(ConverseHistory[RealmName]) do
			pfx = string.sub(key,1,2);
			cname = string.sub(key, 3);
			if (pfx == CONVERSE_PREFIX_MSG) and not lfriends[cname] then
				ConverseHistory[RealmName][key] = nil;
			end
		end
		cMessage(CONVERSE_COLOUR_PRINT..CONVERSE_NAME..": "..CONVERSE_MSG_CLEARRAND..RealmName);
	else
		return;
		-- Huh?!??
	end

	-- Update;
	ConverseItems_Rebuild();
	ConverseItems_Update();
	if (cNumItems > 0) and (cSelected > 0) then
		ConverseHistory_GenerateText(cItems[cSelected].key);
	else
		cText = "";
	end
	ConverseHistory_Update();
end

--=== The guts =============================================================--

function Converse_OnEvent()
	local cname, lname;
	local pname;

	cChannelKludge = "";

	-- You never know if someone has gotten kicked...
	--GuildName = GetGuildInfo("player");
	--if (not GuildName) then
	--	GuildName = "";
	--end

	if (event == "VARIABLES_LOADED") then
		Converse_InitData();
		Converse_InitState();
	elseif (event == "GUILD_ROSTER_UPDATE") then
		-- And more kludginess
		Converse_GuildKludge()

		-- Refresh
		ConverseItems_Rebuild();
		ConverseItems_Update();
		ConverseHistory_Update();
	elseif (event == "CHAT_MSG_CHANNEL_NOTICE") then
		lname = string.lower(string.sub(arg9, string.find(arg9,"[^ ]*")));
		if (arg1 == "YOU_LEFT") then
			cChannelKludge = lname;
			cChannelNames[lname] = nil;
			if ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..lname] and 
			   table.getn(ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..lname].text) == 0 then
				ConverseHistory[RealmName][CONVERSE_PREFIX_CHANNEL..lname] = nil;
			end
		elseif (arg1 == "YOU_JOINED") then
			-- Try to build a list of names for this channel, if it's not a server channel
			Converse_RefreshChannelsList();

			if not cServerChannels[lname] then
				cChannelNames[lname] = {};
				ListChannelByName(lname);
			end
		end
		ConverseItems_Rebuild();
		ConverseItems_Update();
		ConverseHistory_Update();
	elseif (event == "CHAT_MSG_CHANNEL_NOTICE_USER") then
		lname = string.lower(string.sub(arg9, string.find(arg9,"[^ ]*")));
		--cMessage(CONVERSE_COLOUR_PRINT..event..": "..lname..", arg1="..arg1..", arg2="..arg2..", arg5="..arg5);
		if (arg1 == "SET_VOICE") then
			if cChannelNames[lname] then cChannelNames[lname][arg2] = CONVERSE_TAG_NORMAL; end
		elseif (arg1 == "UNSET_VOICE") then
			if cChannelNames[lname] then cChannelNames[lname][arg2] = CONVERSE_TAG_MUTE; end
		elseif (arg1 == "SET_MODERATOR") then
			if cChannelNames[lname] then cChannelNames[lname][arg2] = CONVERSE_TAG_MODERATOR; end
		elseif (arg1 == "UNSET_MODERATOR") then
			if cChannelNames[lname] then cChannelNames[lname][arg2] = CONVERSE_TAG_NORMAL; end
		elseif (arg1 == "CHANNEL_OWNER") or (arg1 == "OWNER_CHANGED") then
			ListChannelByName(lname);	-- We know new owner, but not who lost owner. So, rebuild :(
		end
		ConverseItems_Rebuild();
		ConverseItems_Update();
		if not cPaused then
			ConverseHistory_Update();
		end
	elseif (event == "FRIENDLIST_UPDATE") then
		ConverseItems_Rebuild();
		ConverseItems_Update();
	elseif (event == "CHAT_MSG_CHANNEL_LIST") then
		if arg4 and string.find(arg4,"%a[%a%d]+") then
			lname = string.lower(string.sub(arg4, string.find(arg4,"%a[%a%d]+"))); -- Channel name
			if not cServerChannels[lname] then
				if not cChannelNames[lname] then cChannelNames[lname] = {}; end
				-- arg1 is a comma-delimited list of names. Note symbols, "*" for owner, "@" for moderator, "#" for mute
				-- for pname in string.gmatch(arg1, "%a+") do
				for pname in string.gmatch(arg1, "[^, ]+") do
					local tag = CONVERSE_TAG_NORMAL;
					if string.find(pname, "%a+") == 2 then
						tag	= string.sub(pname, 1, 1);
						pname	= string.sub(pname, 2);
					end
					cChannelNames[lname][pname] = tag;
				end
			end
		end
		ConverseLinks_Update();
	else
		-- NOTE: Remember to catch all events first, letting it fallback to chat event when this happens
		if cFinishedLoading then
			Converse_ParseChatMessage(event, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
		end
	end
end

function Converse_GuildKludge()
	if not RealmName or (RealmName == "") then
		-- Why are we hitting this so early?
		RealmName = GetCVar("realmname");
		if not RealmName then
			-- WHAT THE HECK?!?
			--return;
			-- Just so things work, we'll run with a blank RealmName until we can actually get one
			RealmName = "";
		end
	end

	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end

	-- Guild
	if not ConverseHistory[RealmName] then
		ConverseHistory[RealmName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].text = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].buffer then
		ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].buffer = {};
	end
	-- Officer
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName] = {};
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].text = {}; 
	end
	if not ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].buffer then
		ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].buffer = {}; 
	end
end

function Converse_OnShow()
	ConverseState[RealmName][PlayerName].hidden = nil;

	GuildName = GetGuildInfo("player");
	if (not GuildName) then
		GuildName = "";
	end

	-- Button States
	for i=1,CONVERSE_MAXCHANNELS do
		if ConverseState[RealmName][PlayerName].ignore[i] then
			getglobal("ConverseFrameIgnoreCheckButton"..i):SetChecked(1);
		end
	end
	-- And the other 6...
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_SAY] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_SAY):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_PARTY] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_PARTY):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_RAID] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_RAID):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_BG] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_BG):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_GUILD] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_GUILD):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_OFFICER] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_OFFICER):SetChecked(1);
	end
	if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_ID_WHISPER] then
		getglobal("ConverseFrameIgnoreCheckButton"..CONVERSE_IGNORE_ID_WHISPER):SetChecked(1);
	end

	ConverseItems_Rebuild();
	ConverseItems_Update();
	ConverseHistory_Update();

	-- Just for fun
	SetPortraitTexture(ConverseFrameIconTexture, "player");

	ConverseLinks_Update();

	-- AltEdit
	if ConverseState[RealmName][PlayerName].altedit then
		ConverseFrameAltEditFrameText:SetFocus();
		if (cNumItems > 0) and (cSelected > 0) then
			ConverseFrameAltEditFrameText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
		end
	end
end

function Converse_OnHide()
	ConverseState[RealmName][PlayerName].hidden = 1;
end

function Converse_ParseChatMessage(event, text, author, language, channumname, target, flags, zoneid, channum, channame)
	-- Don't you love those args?

	local key;
	local cname;
	local title;
	local oneline;
	local timestamp;
	local i;

	-- By default, this will be...
	if author then
		title = CONVERSE_NAME_PRE..author..CONVERSE_NAME_POST;
	end
	if (event == "CHAT_MSG_WHISPER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
	elseif (event == "CHAT_MSG_WHISPER_INFORM") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		--title = CONVERSE_NAME_PRE..PlayerName..CONVERSE_NAME_REPLY..author..CONVERSE_NAME_POST;
		title = PlayerName..CONVERSE_NAME_REPLY..CONVERSE_NAME_PRE..author..CONVERSE_NAME_POST;
	-- RealID chat
	elseif (event == "CHAT_MSG_BN_WHISPER") then
		--if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		-- Gonna assume you don't want to ignore RealID stuff... at least for now so I don't have to code it
		key = CONVERSE_PREFIX_MSGRID..author;
	elseif (event == "CHAT_MSG_BN_WHISPER_INFORM") then
		--if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		-- Gonna assume you don't want to ignore RealID stuff... at least for now so I don't have to code it
		key = CONVERSE_PREFIX_MSGRID..author;
		--title = CONVERSE_NAME_PRE..PlayerName..CONVERSE_NAME_REPLY..author..CONVERSE_NAME_POST;
		title = PlayerName..CONVERSE_NAME_REPLY..CONVERSE_NAME_PRE..author..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_AFK") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		title = author..CONVERSE_NAME_AFK;
	elseif (event == "CHAT_MSG_DND") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then return; end
		key = CONVERSE_PREFIX_MSG..author;
		title = author..CONVERSE_NAME_DND;
	elseif (event == "CHAT_MSG_GUILD") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then return; end
		key = CONVERSE_PREFIX_GUILD..GuildName;
	elseif (event == "CHAT_MSG_OFFICER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then return; end
		key = CONVERSE_PREFIX_OFFICER..GuildName;
	elseif (event == "CHAT_MSG_PARTY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then return; end
		key = CONVERSE_INDEX_PARTY;
	elseif (event == "CHAT_MSG_PARTY_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then return; end
		key = CONVERSE_INDEX_PARTY;
		title = CONVERSE_NAME_PLEADPRE..author..CONVERSE_NAME_POST;
	-- Say, yell, and all that fun stuff
	elseif (event == "CHAT_MSG_SAY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_SAY;
	elseif (event == "CHAT_MSG_YELL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_YELL;
	elseif (event == "CHAT_MSG_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author.." ";
	elseif (event == "CHAT_MSG_TEXT_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = "";
	-- Monsters!
	elseif (event == "CHAT_MSG_MONSTER_SAY") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_SAY;
	elseif (event == "CHAT_MSG_MONSTER_YELL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author..CONVERSE_NAME_YELL;
	elseif (event == "CHAT_MSG_MONSTER_EMOTE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then return; end
		key = CONVERSE_INDEX_SAY;
		title = author.." ";
	-- Raid, Raid leader, Raid warning
	elseif (event == "CHAT_MSG_RAID") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
	elseif (event == "CHAT_MSG_RAID_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
		title = CONVERSE_NAME_RLEADPRE..author..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_RAID_WARNING") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then return; end
		key = CONVERSE_INDEX_RAID;
		title = CONVERSE_NAME_RWARNPRE..author..CONVERSE_NAME_POST;
	-- Battleground, Battleground leader, and in-BG messages
	elseif (event == "CHAT_MSG_BATTLEGROUND") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then return; end
		key = CONVERSE_INDEX_BG;
	elseif (event == "CHAT_MSG_BATTLEGROUND_LEADER") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then return; end
		key = CONVERSE_INDEX_BG;
		title = CONVERSE_NAME_BLEADPRE..author..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_BG_SYSTEM_ALLIANCE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then return; end
		key = CONVERSE_INDEX_BG;
		title = cChatColourString("BG_SYSTEM_ALLIANCE");
		text = text.."|r";
	elseif (event == "CHAT_MSG_BG_SYSTEM_HORDE") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then return; end
		key = CONVERSE_INDEX_BG;
		title = cChatColourString("BG_SYSTEM_HORDE");
		text = text.."|r";
	elseif (event == "CHAT_MSG_BG_SYSTEM_NEUTRAL") then
		if ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then return; end
		key = CONVERSE_INDEX_BG;
		title = cChatColourString("BG_SYSTEM_NEUTRAL");
		text = text.."|r";
	-- Channel
	elseif (event == "CHAT_MSG_CHANNEL") then
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		title = CONVERSE_NAME_PRE..author..CONVERSE_NAME_POST;
	elseif (event == "CHAT_MSG_CHANNEL_JOIN") then
		--key = CONVERSE_PREFIX_CHANNEL..string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		---
		if not cServerChannels[cname] and cChannelNames[cname] then
			if not cChannelNames[cname][author] then
				cChannelNames[cname][author] = CONVERSE_TAG_NORMAL;
			end
		end
		----
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		title = author..CONVERSE_NAME_JOIN;
		text = "";
	elseif (event == "CHAT_MSG_CHANNEL_LEAVE") then
		--key = CONVERSE_PREFIX_CHANNEL..string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		cname = string.lower(string.sub(channame, string.find(channame,"[^ ]*")));
		key = CONVERSE_PREFIX_CHANNEL..cname;
		---
		if not cServerChannels[cname] and cChannelNames[cname] then
			if cChannelNames[cname][author] then
				cChannelNames[cname][author] = nil;
			end
		end
		----
		if ConverseState[RealmName][PlayerName].ignore[channum] then return; end
		title = author..CONVERSE_NAME_LEAVE;
		text = "";
	elseif (event == "GUILD_MOTD") then
		if string.len(arg1) > 0 then
			key = CONVERSE_PREFIX_GUILD..GuildName;
			title = "";
			text = string.format(TEXT(GUILD_MOTD_TEMPLATE), arg1);
		else
			return;
		end
	else
		-- SHOULD NOT GET THIS!
		return;
	end
	-- Filter out GEM messages
	if ConverseState[RealmName][PlayerName].gemfilter then
		if string.find(text,"<GEM%d+.%d+") == 1 then
			return;
		end
	end

	timestamp = string.sub(date(),10,17);


	-- Does this item exist in the history?
	if not ConverseHistory[RealmName][key] then
		ConverseHistory[RealmName][key] = {};
		if ConverseState[RealmName][PlayerName].datestamp then
			local dateline = cChatColourString(string.sub(event,10), true)..cDateStamp.."|r";
			ConverseHistory[RealmName][key].text = {dateline};
			ConverseHistory[RealmName][key].datestamp = cDateStamp;
			ConverseHistory[RealmName][key].buffer = {};
		else
			ConverseHistory[RealmName][key].text = {};
			ConverseHistory[RealmName][key].buffer = {};
		end
	else
		-- Check if we're adding a datestamp, or replacing a datestamp
		if ConverseState[RealmName][PlayerName].datestamp and
		   ((not ConverseHistory[RealmName][key].datestamp) or (ConverseHistory[RealmName][key].datestamp ~= cDateStamp)) then
			local dateline = "";
			if (string.sub(key,1,2) == CONVERSE_PREFIX_CHANNEL) then
				-- Account for colour numbers
				if cname and cChannels[cname] then
					dateline = cChatColourString("CHANNEL"..cChannels[cname], true);
				end
			else
				dateline = cChatColourString(string.sub(event,10), true);
			end
			dateline = dateline..cDateStamp.."|r";

			ConverseHistory[RealmName][key].datestamp = cDateStamp;
			table.insert(ConverseHistory[RealmName][key].text, dateline)

			-- Add to cText if current, so we don't have to reparse .text
			if (key == ConverseState[RealmName][PlayerName].lastselected) then
				cText = cText..dateline.."\n";
				cLines = cLines + 1;
			end		
		end
	end


	-- Check if we've hit the line maximum
	local numlines = table.getn(ConverseHistory[RealmName][key].text);
	if (ConverseState[RealmName][PlayerName].maxlines > 0) and 
	   (numlines > ConverseState[RealmName][PlayerName].maxlines) then
		table.remove(ConverseHistory[RealmName][key].text, numlines - ConverseState[RealmName][PlayerName].maxlines + 1);
	end

	-- Build the line
	oneline = title..text;

	-- Colour Mangling
	if (event == "CHAT_MSG_EMOTE") or
	   (event == "CHAT_MSG_TEXT_EMOTE") or
	   (event == "CHAT_MSG_YELL") or
	   (event == "CHAT_MSG_MONSTER_SAY") or
	   (event == "CHAT_MSG_MONSTER_EMOTE") or
	   (event == "CHAT_MSG_MONSTER_YELL") then
		local chattype = string.sub(event, 10);
		local linecolour = cChatColourString(chattype);
		oneline = linecolour..oneline;

		-- Look for |h|r's in the string, and append the colour again, otherwise it'll revert to the 'Say' colour
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);

		-- And the trailing colour return
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_CHANNEL_JOIN") or (event == "CHAT_MSG_CHANNEL_LEAVE") then
		oneline = CONVERSE_COLOUR_GREY..oneline.."|r";
	elseif (event == "CHAT_MSG_PARTY_LEADER") then
		local linecolour = cChatColourString("PARTY_LEADER");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_RAID_LEADER") then
		local linecolour = cChatColourString("RAID_LEADER");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_RAID_WARNING") then
		local linecolour = cChatColourString("RAID_WARNING");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	elseif (event == "CHAT_MSG_BATTLEGROUND_LEADER") then
		local linecolour = cChatColourString("BATTLEGROUND_LEADER");
		oneline = linecolour..oneline;
		oneline = string.gsub(oneline, "|h|r", "|h|r"..linecolour);
		oneline = oneline.."|r";
	end

	-- Timestamp
	oneline = timestamp.."| "..oneline;

	-- MWBIGChat hook
	if MWBIGChat then
		MWBIGChat(event, text, author, language, channumname, target, flags, zoneid, channum, channame)
	end

	-- DONE!

	-- If we are paused, add it to the buffer, else add it to the history
	if cPaused then
		table.insert(ConverseHistory[RealmName][key].buffer, oneline)
	else
		table.insert(ConverseHistory[RealmName][key].text, oneline)

		-- DEBUG: cMessage("CM | key = "..key.." | oneline = "..oneline);

		-- if current item, update cText
		if (key == ConverseState[RealmName][PlayerName].lastselected) then
			--ConverseHistory_GenerateText(key)
			cText = cText..oneline.."\n";
			cLines = cLines + 1;
			ConverseHistory_Update();
		end
	end

	-- Required for flashing...
	-- No activity on channel joins/leaves
	if (event ~= "CHAT_MSG_CHANNEL_JOIN") and
	   (event ~= "CHAT_MSG_CHANNEL_LEAVE") and
	   (event ~= "CHAT_MSG_WHISPER_INFORM") then
	   	if not ConverseHistory[RealmName][key].activity then
			ConverseHistory[RealmName][key].activity = 1;
		else
			ConverseHistory[RealmName][key].activity = ConverseHistory[RealmName][key].activity + 1;
		end
		ConverseHistory[RealmName][key].lastdate = date();
	end

	ConverseItems_Rebuild();

	ConverseItems_Update();
	ConverseLinks_Update();
end

--=== Stupid FauxScrollBars! ===============================================--

function ConverseItems_Rebuild()
	local key, value;
	local i, j;
	local nametable = {};
	local pfx, cname, lname;

	-- KLUDGE
	Converse_GuildKludge();

	-- Friends experimental
	local lfriends = {};
	for i=1,GetNumFriends() do
		local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
		if not fname then
			--DEFAULT_CHAT_FRAME:AddMessage("NO FNAME i="..i);
		else
			if fconn then
				lfriends[fname] = 1;
			else
				lfriends[fname] = 0;
			end
		end
	end
	
	-- RealID friends online/offline
	local lrid = {};
	local bncount,bncountonline = BNGetNumFriends();
	for i=1,bncount do
		local presenceID, givenName, surname, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, messageText, noteText = BNGetFriendInfo(i);
		local key=givenName.." "..surname
		if isOnline then
			lrid[key] = 1;
		else
			lrid[key] = 0;
		end
	end

	-- Default 4
	i = 1;
	cItems = {};	-- This is our table of items in the lefthand pane of the window

	-- Kludge for certain events triggering this function before certain variables are available...
	if not RealmName then RealmName=""; end
	if not PlayerName then PlayerName=UnitName("player"); end
	if not ConverseState[RealmName][PlayerName] then ConverseState[RealmName][PlayerName] = {}; end
	if not ConverseState[RealmName][PlayerName].ignore then ConverseState[RealmName][PlayerName].ignore = {}; end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_SAY] then
		cItems[i]		= {};
		cItems[i].key 		= CONVERSE_INDEX_SAY;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["SAY"].r;
		cItems[i].colour.g	= ChatTypeInfo["SAY"].g;
		cItems[i].colour.b	= ChatTypeInfo["SAY"].b;
		cItems[i].name		= CONVERSE_LABEL_SAY;
		if ConverseHistory[RealmName][CONVERSE_INDEX_SAY].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_PARTY] then
		cItems[i]		= {};
		cItems[i].key 		= CONVERSE_INDEX_PARTY;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["PARTY"].r;
		cItems[i].colour.g	= ChatTypeInfo["PARTY"].g;
		cItems[i].colour.b	= ChatTypeInfo["PARTY"].b;
		cItems[i].name		= CONVERSE_LABEL_PARTY;
		if ConverseHistory[RealmName][CONVERSE_INDEX_PARTY].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_RAID] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_INDEX_RAID;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["RAID"].r;
		cItems[i].colour.g	= ChatTypeInfo["RAID"].g;
		cItems[i].colour.b	= ChatTypeInfo["RAID"].b;
		cItems[i].name		= CONVERSE_LABEL_RAID;
		if ConverseHistory[RealmName][CONVERSE_INDEX_RAID].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_BG] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_INDEX_BG;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["BATTLEGROUND"].r;
		cItems[i].colour.g	= ChatTypeInfo["BATTLEGROUND"].g;
		cItems[i].colour.b	= ChatTypeInfo["BATTLEGROUND"].b;
		cItems[i].name		= CONVERSE_LABEL_BG;
		if ConverseHistory[RealmName][CONVERSE_INDEX_BG].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_GUILD] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_PREFIX_GUILD..GuildName;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["GUILD"].r;
		cItems[i].colour.g	= ChatTypeInfo["GUILD"].g;
		cItems[i].colour.b	= ChatTypeInfo["GUILD"].b;
		cItems[i].name		= CONVERSE_LABEL_GUILD.." <"..GuildName..">";
		if ConverseHistory[RealmName][CONVERSE_PREFIX_GUILD..GuildName].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_OFFICER] then
		cItems[i]		= {};
		cItems[i].key		= CONVERSE_PREFIX_OFFICER..GuildName;
		cItems[i].colour	= {};
		cItems[i].colour.r	= ChatTypeInfo["OFFICER"].r;
		cItems[i].colour.g	= ChatTypeInfo["OFFICER"].g;
		cItems[i].colour.b	= ChatTypeInfo["OFFICER"].b;
		cItems[i].name		= CONVERSE_LABEL_OFFICER.." <"..GuildName..">";
		if ConverseHistory[RealmName][CONVERSE_PREFIX_OFFICER..GuildName].activity then cItems[i].activity = 1; end
		i = i + 1;
	end

	-- Channels and Private messages
	Converse_RefreshChannelsList();
	for key, value in pairs(ConverseHistory[RealmName]) do
		--if ((key ~= CONVERSE_INDEX_PARTY) and
		--    (key ~= CONVERSE_INDEX_RAID) and
		--    (string.sub(key,1,2) ~= CONVERSE_PREFIX_GUILD) and
		--    (string.sub(key,1,2) ~= CONVERSE_PREFIX_OFFICER)
		--   ) then
		pfx = string.sub(key,1,2);
		cname = string.sub(key, 3);
		lname = string.lower(cname);
		if (pfx == CONVERSE_PREFIX_MSG) then
			-- Private Message: "m/PlayerName"
		   	-- Assuming blizzard doesn't allow 1- or 0-letter names in the future...
			if not ConverseState[RealmName][PlayerName].ignore[CONVERSE_IGNORE_WHISPER] then
				if lfriends[cname] == 1 then
					-- Friends we substitue "k/" in front so they end up sorted ahead of other messages "m/"
					table.insert(nametable, CONVERSE_PREFIX_MSGFRIEND..cname);
				elseif lfriends[cname] == 0 then
					-- "l/" for offline friends next
					table.insert(nametable, CONVERSE_PREFIX_MSGFRIENDOFFLINE..cname);
				else
					table.insert(nametable, key);
				end
			end
		elseif (pfx == CONVERSE_PREFIX_MSGRID) then
			if lrid[cname] == 1 then
				table.insert(nametable, CONVERSE_PREFIX_MSGRID..cname);
			elseif lrid[cname] == 0 then
				table.insert(nametable, CONVERSE_PREFIX_MSGRIDOFFLINE..cname);
			else
				table.insert(nametable, CONVERSE_PREFIX_MSGRID..cname);
			end
		elseif (pfx == CONVERSE_PREFIX_CHANNEL) then
			-- Channels: "c/##/Channelname"
			-- We tack on the number so we can sort by number
			if cChannels[lname] then
				if not ConverseState[RealmName][PlayerName].ignore[cChannels[lname]] then
					table.insert(nametable, CONVERSE_PREFIX_CHANNEL..string.format("%02d", cChannels[lname])..lname);
				end
			else
				-- Unnumbered channels (ie. currently not listening to) come last.
				table.insert(nametable, CONVERSE_PREFIX_CHANNEL.."99"..lname);
			end
		end
	end
	table.sort(nametable);

	for j=1,table.getn(nametable) do
		--cMessage(j.." = '"..nametable[j].."'");
		cItems[i] = {};
		pfx = string.sub(nametable[j],1,2)
		if pfx == CONVERSE_PREFIX_CHANNEL then
			-- Remove the extraneous number which was there only for sorting.
			cname = string.sub(nametable[j],5);
			lname = string.lower(cname);
			cItems[i].key		= CONVERSE_PREFIX_CHANNEL..lname;
			cItems[i].colour	= {};
			if cChannels[lname] then
				local _, fullname = GetChannelName(cChannels[lname]);
				cItems[i].name	= cChannels[lname]..". "..fullname;
				-- Experimental:
				cItems[i].colour["r"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].r;
				cItems[i].colour["g"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].g; 
				cItems[i].colour["b"]	= ChatTypeInfo["CHANNEL"..cChannels[lname]].b; 
			else
				cItems[i].name	= lname;
				cItems[i].colour["r"]	= CONVERSE_COLOUR_CHANNEL["r"];
				cItems[i].colour["g"]	= CONVERSE_COLOUR_CHANNEL["g"];
				cItems[i].colour["b"]	= CONVERSE_COLOUR_CHANNEL["b"];
			end
		elseif (pfx == CONVERSE_PREFIX_MSGRID) or (pfx == CONVERSE_PREFIX_MSGRIDOFFLINE) then
			-- RealID message
			cname = string.sub(nametable[j], 3);
			-- Substitute "i/" back for "j/" so we get the right index
			cItems[i].key		= nametable[j];
			if (pfx == CONVERSE_PREFIX_MSGRIDOFFLINE) then
				cItems[i].key	= CONVERSE_PREFIX_MSGRID..cname;
			end
			cItems[i].colour	= {};
			cItems[i].colour.r	= ChatTypeInfo["BN_WHISPER"].r;
			cItems[i].colour.g	= ChatTypeInfo["BN_WHISPER"].g;
			cItems[i].colour.b	= ChatTypeInfo["BN_WHISPER"].b;
			if lrid[cname] == 1 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_ONLINE;
			elseif lrid[cname] == 0 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_OFFLINE;
			else
				cItems[i].name		= cname;
			end
		--elseif (pfx == CONVERSE_PREFIX_MSGFRIEND) or (pfx == CONVERSE_PREFIX_MSGFRIENDOFFLINE) then
		else
			-- Private messages
			cname = string.sub(nametable[j], 3);
			-- Substitute "m/" back for "l/" so we get the right index
			cItems[i].key		= nametable[j];
			if (pfx == CONVERSE_PREFIX_MSGFRIEND) or (pfx == CONVERSE_PREFIX_MSGFRIENDOFFLINE) then
				cItems[i].key	= CONVERSE_PREFIX_MSG..cname;
			end
			cItems[i].colour	= {};
			cItems[i].colour.r	= ChatTypeInfo["WHISPER"].r;
			cItems[i].colour.g	= ChatTypeInfo["WHISPER"].g;
			cItems[i].colour.b	= ChatTypeInfo["WHISPER"].b;
			if lfriends[cname] == 1 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_ONLINE;
			elseif lfriends[cname] == 0 then	
				cItems[i].name		= cname.." "..CONVERSE_NAME_OFFLINE;
			else
				cItems[i].name		= cname;
			end
		end
		--cMessage(j..": KEY = '"..cItems[i].key.."'");
		-- Activity tag
		if ConverseHistory[RealmName][cItems[i].key].activity then
			cItems[i].activity = 1;
		else
			cItems[i].activity = nil;
		end

		i = i + 1;
	end

	cNumItems = (i - 1);

	cSelected = -1;
	for i=1,table.getn(cItems) do
		if (cItems[i].key == ConverseState[RealmName][PlayerName].lastselected) then
			cSelected = i;
		end
	end

	if (cSelected == -1) and (cNumItems > 0) then
		-- Okay, we didn't find it. Got cleared or ignored
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
	end
end

-- Items buttons
function ConverseItems_OnClick()
	local id = this:GetID();
	button = getglobal("ConverseFrameItemsButton"..id);

	cSelected = button.listindex;

	if not cItems[cSelected] then
		-- Uh oh, we have a stray item that has no name...
		ConverseItems_Rebuild();
		ConverseItems_Update();
		return;
	end

	ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;

	-- Turn off flashing
	if not cPaused then
		ConverseHistory[RealmName][cItems[cSelected].key].activity = nil;
		cItems[cSelected].activity = nil;
	end

	-- Change history to match
	ConverseHistory_GenerateText(cItems[cSelected].key);
	ConverseHistory_Update();

	-- Set the 'selected' text.
	ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
	ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	ConverseItems_Update();

	-- AltEdit
	if ConverseState[RealmName][PlayerName].altedit then
		ConverseFrameAltEditFrameText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	end
end

function ConverseItems_Update()
	local index;
	local offset = FauxScrollFrame_GetOffset(ConverseFrameItemsScrollFrame);
	local button, buttontext;

	for i=1, CONVERSE_ITEMS_NUMSHOWN do
		index = i + offset;
		button = getglobal("ConverseFrameItemsButton"..i);
		button.listindex = index;

		buttontext = getglobal("ConverseFrameItemsButton"..i.."Text");
		if (index <= cNumItems) then
			buttontext:SetText(cItems[index].name);
			buttontext:SetTextColor(cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
			if (cItems[index].activity) then
				if ConverseState[RealmName][PlayerName].blink or
				   (ConverseState[RealmName][PlayerName].chanblink and string.sub(cItems[index].key,1,2) == CONVERSE_PREFIX_CHANNEL) or
				   (ConverseState[RealmName][PlayerName].msgblink and string.sub(cItems[index].key,1,2) == CONVERSE_PREFIX_MSG) or
				   (ConverseState[RealmName][PlayerName].msgblink and string.sub(cItems[index].key,1,2) == CONVERSE_PREFIX_MSGRID) then
					UIFrameFlash(button, 0.5, 0.5, -1, true, 0, 0.25);
				else
					UIFrameFlashStop(button);
				end
			else
				UIFrameFlashStop(button);
			end
			button:Show();
		else
			button:UnlockHighlight();
			buttontext:SetText("");
			UIFrameFlashStop(button);
			button:Hide();
		end

		if (cNumItems > 0) and (index == cSelected) then
			cItems[index].activity = nil;
			ConverseHistory[RealmName][cItems[index].key].activity = nil;

			button:LockHighlight();
			UIFrameFlashStop(button);
			button:Show();
		else
			button:UnlockHighlight();
		end
	end

	FauxScrollFrame_Update(ConverseFrameItemsScrollFrame, cNumItems, CONVERSE_ITEMS_NUMSHOWN, CONVERSE_ITEMS_HEIGHT);

	if (cNumItems > 0) and (cSelected > 0) then
		ConverseFrameItemsSelectedText:SetText(cItems[cSelected].name);
		ConverseFrameItemsSelectedText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	else
		ConverseFrameItemsSelectedText:SetText("");
	end

	-- KLUDGE: Since I don't have a graphical divider other than the scrollbar
	ConverseFrameItemsScrollFrame:Show();

	-- Links/Players frame
	if (cNumItems > 0) and (cSelected > 0) and
	   (string.sub(cItems[cSelected].key,1,2) == CONVERSE_PREFIX_CHANNEL) and not cServerChannels[string.sub(cItems[cSelected].key,3)] then
		ConverseFrameLinksText:SetText(CONVERSE_HEADER_PLAYERS);
	else
		ConverseFrameLinksText:SetText(CONVERSE_HEADER_LINKS);
	end

	ConverseLinks_Update();
end

function ConverseItems_OnEnter()
	local index	= this:GetID();

	if not cItems[index] then
		-- Uh oh, we have a stray item that has no name...
		ConverseItems_Rebuild();
		ConverseItems_Update();
		return;
	end

	local key	= cItems[index].key;
	local prefix	= string.sub(cItems[index].key,1,2);
	local cname	= string.sub(cItems[index].key,3);
	local lname	= string.lower(cname);
	local newlines	= 0;

	local numraid	= GetNumRaidMembers();
	local numparty	= GetNumPartyMembers() + 1;	-- Nice, it's 0-4!

	GameTooltip:SetOwner(getglobal("ConverseFrameItemsButton"..index), "ANCHOR_BOTTOMRIGHT");
	if (prefix == CONVERSE_PREFIX_MSG) then
		-- Show location of friend
		-- FIXME: I should really have a global friends structure instead of redoing it every time...
		local locfriends = {};
		local i;
		for i=1,GetNumFriends() do
			local fname, flevel, fclass, farea, fconn, fstat = GetFriendInfo(i);
			if fname and fconn then
				if (fstat ~= "") then
					locfriends[fname] = farea.." "..fstat;
				else
					locfriends[fname] = farea;
				end
			end
		end

		if locfriends[cname] then
			GameTooltip:AddDoubleLine(cItems[index].name,
			                          locfriends[cname],
			                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
						  CONVERSE_COLOUR_TOOLTIP_LOCATION.r, CONVERSE_COLOUR_TOOLTIP_LOCATION.g, CONVERSE_COLOUR_TOOLTIP_LOCATION.b);
		else
			GameTooltip:AddLine(cItems[index].name, cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
		end
	elseif (prefix == CONVERSE_PREFIX_MSGRID) then
		-- RealID junk
		local locrid = {};
		local bncount, bncountonline = BNGetNumFriends();
		for i=1,bncount do
			local presenceID, givenName, surname, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, messageText, noteText = BNGetFriendInfo(i);
			local key=givenName.." "..surname;
			local lstat="";
			if isOnline then
				if isAFK then
					lstat="<Away>"
				elseif isDND then
					lstat="<Busy>"
				end
				if client then
					if toonName then
						if toonID then
							local hasFocus, toonName, client, realmName, faction, race, class, guild, zoneName, level, gameText = BNGetToonInfo(toonID);
							if client == "WoW" then
								locrid[key]=toonName.." / "..zoneName;
							else
								locrid[key]=toonName.." ("..client..")";
							end
						else
							locrid[key]=toonName.." ("..client..")";
						end
					else
						locrid[key]="("..client..")";
					end
				end
			end
		end

		if locrid[cname] then
			GameTooltip:AddDoubleLine(cItems[index].name,
			                          locrid[cname],
			                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
						  CONVERSE_COLOUR_TOOLTIP_LOCATION.r, CONVERSE_COLOUR_TOOLTIP_LOCATION.g, CONVERSE_COLOUR_TOOLTIP_LOCATION.b);
		else
			GameTooltip:AddLine(cItems[index].name, cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
		end
	elseif (prefix == CONVERSE_PREFIX_CHANNEL) and not cServerChannels[lname] and cChannelNames[lname] then
		-- Show number of players in channel
		local i,v, numplayers;
		numplayers = 0;
		for i,v in pairs(cChannelNames[lname]) do
			numplayers = numplayers + 1;
		end
		GameTooltip:AddDoubleLine(cItems[index].name,
		                          string.format(CONVERSE_TOOLTIP_FORMAT_PLAYERS, numplayers),
		                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
		                          CONVERSE_COLOUR_TOOLTIP_PLAYERS.r, CONVERSE_COLOUR_TOOLTIP_PLAYERS.g, CONVERSE_COLOUR_TOOLTIP_PLAYERS.b);
	elseif (key == CONVERSE_INDEX_PARTY) and (numparty > 0) then
		GameTooltip:AddDoubleLine(cItems[index].name,
		                          string.format(CONVERSE_TOOLTIP_FORMAT_PARTY, numparty),
		                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
		                          CONVERSE_COLOUR_TOOLTIP_PLAYERS.r, CONVERSE_COLOUR_TOOLTIP_PLAYERS.g, CONVERSE_COLOUR_TOOLTIP_PLAYERS.b);
	elseif (key == CONVERSE_INDEX_RAID) and (numraid > 0) then
		GameTooltip:AddDoubleLine(cItems[index].name,
		                          string.format(CONVERSE_TOOLTIP_FORMAT_RAID, numraid),
		                          cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b,
		                          CONVERSE_COLOUR_TOOLTIP_PLAYERS.r, CONVERSE_COLOUR_TOOLTIP_PLAYERS.g, CONVERSE_COLOUR_TOOLTIP_PLAYERS.b);
	else
		-- Fallback
		GameTooltip:AddLine(cItems[index].name, cItems[index].colour.r, cItems[index].colour.g, cItems[index].colour.b);
	end


	-- Show number of lines
	if ConverseHistory[RealmName][key].activity then
		newlines = ConverseHistory[RealmName][key].activity;
	end

	GameTooltip:AddLine(string.format(CONVERSE_TOOLTIP_FORMAT_LINES,
	                    table.getn(ConverseHistory[RealmName][key].text),
	                    ConverseState[RealmName][PlayerName].maxlines, newlines),
			    CONVERSE_COLOUR_TOOLTIP_LINES.r, CONVERSE_COLOUR_TOOLTIP_LINES.g, CONVERSE_COLOUR_TOOLTIP_LINES.b , 1)

	if ConverseHistory[RealmName][key].lastdate then
		GameTooltip:AddLine(string.format(CONVERSE_TOOLTIP_FORMAT_LASTDATE, ConverseHistory[RealmName][key].lastdate),
		                    CONVERSE_COLOUR_TOOLTIP_LASTDATE.r, CONVERSE_COLOUR_TOOLTIP_LASTDATE.g, CONVERSE_COLOUR_TOOLTIP_LASTDATE.b, 1);
	end

	GameTooltip:Show();
end

--=== History ==============================================================--

function ConverseHistory_SetFontSize(size)
	local fontfile, _, fontflags;

	-- History frame
	fontfile, _, fontflags = ConverseFrameHistoryScrollFrameText:GetFont();
	ConverseFrameHistoryScrollFrameText:SetFont(fontfile, size, fontflags);
	ConverseHistory_Update();

	-- Links frame
	fontfile, _, fontflags = ConverseFrameLinksPanelMessageFrame:GetFont();
	ConverseFrameLinksPanelMessageFrame:SetFont(fontfile, size, fontflags);

	-- Multipaste frame
	fontfile, _, fontflags = ConverseFrameMultipasteFrameScrollFrameText:GetFont();
	ConverseFrameMultipasteFrameScrollFrameText:SetFont(fontfile, size, fontflags);

	ConverseLinks_Update();
end

function ConverseHistory_Clear()
	-- if (IsShiftKeyDown()) then
		local prefix	= string.sub(cItems[cSelected].key, 1, 2);
		local lname	= string.lower(string.sub(cItems[cSelected].key, 3));
		-- If it's a player or a channel we are not in anymore, remove them from the list
		if (prefix == CONVERSE_PREFIX_MSG) or (prefix == CONVERSE_PREFIX_MSGRID) or
		   ((prefix == CONVERSE_PREFIX_CHANNEL) and (not cChannels[lname])) then
			ConverseHistory[RealmName][cItems[cSelected].key] = nil;

			-- Reset to default item
			ConverseItems_Rebuild();

			cSelected = 1;
			if (cNumItems > 0) then
				ConverseState[RealmName][PlayerName].lastselected = cItems[cSelected].key;
				ConverseHistory_GenerateText(ConverseState[RealmName][PlayerName].lastselected);
			else
				cText = "";
				cLines = 0;
			end

			ConverseItems_Update();
			ConverseHistory_Update();
		else
			ConverseHistory[RealmName][cItems[cSelected].key].text = {};
			ConverseHistory[RealmName][cItems[cSelected].key].datestamp = nil;
			cText = "";
			cLines = 0;
			ConverseFrameHistoryScrollFrameText:SetText("");
		end

	ConverseLinks_Update();
	-- end
end

function ConverseHistory_PauseResume()
	if not cPaused then
		cPaused = 1;

		ConverseFrameTitleText:SetText(CONVERSE_HEADER_TITLE.." ("..CONVERSE_VER..") - "..RealmName.." "..CONVERSE_HEADER_PAUSED);

		ConverseFramePauseButtonText:SetText(CONVERSE_BUTTON_RESUME);
		UIFrameFlash(ConverseFramePauseButtonText, 0.5, 0.5, -1, true, 0, 0.25);
	else
		cPaused = nil;

		ConverseFrameTitleText:SetText(CONVERSE_HEADER_TITLE.." ("..CONVERSE_VER..") - "..RealmName);

		ConverseFramePauseButtonText:SetText(CONVERSE_BUTTON_PAUSE);
		UIFrameFlashStop(ConverseFramePauseButtonText);

		-- Add the buffered lines to the history
		ConverseHistory_UpdateFromBuffers();

		-- Update;
		ConverseItems_Rebuild();
		ConverseItems_Update();
		if (cNumItems > 0) and (cSelected > 0) then
			ConverseHistory_GenerateText(cItems[cSelected].key);
		else
			cText = "";
		end
		ConverseHistory_Update();
	end
end

function ConverseHistory_GenerateText(key)
	local i;
	cText = "";
	cLines = 0;
	if (cNumItems > 0) and (cSelected > 0) and ConverseHistory[RealmName][key] then
		if ConverseHistory[RealmName][key] and ConverseHistory[RealmName][key].text then
			for i = 1, table.getn(ConverseHistory[RealmName][key].text) do
				cText = cText..ConverseHistory[RealmName][key].text[i].."\n";
				cLines = cLines + 1;
			end
		end
	end
end

--function ConverseHistory_ToggleLock()
--	if ConverseState[RealmName][PlayerName].historylock then
--		ConverseState[RealmName][PlayerName].historylock = nil;
--		ConverseFrameSaveButton:Enable();
--	else
--		ConverseState[RealmName][PlayerName].historylock = 1;
--		ConverseFrameSaveButton:Disable();
--	end
--	ConverseHistory_Update();
--end

function ConverseHistory_OnTextChanged()
	--if ConverseState[RealmName][PlayerName].historylock then
		-- We are not allowing changes
		if ConverseFrameHistoryScrollFrameText:GetText() ~= cText then
			ConverseFrameHistoryScrollFrameText:SetText(cText);
			ConverseFrameHistoryScrollFrameText:ClearFocus();
		end
	--end
	ConverseHistory_UpdateScrollbar();
end

function ConverseHistory_UpdateScrollbar()
	ConverseFrameHistoryScrollFrame:UpdateScrollChildRect();
	local min, max = ConverseFrameHistoryScrollFrameScrollBar:GetMinMaxValues();
	if (max > 0) and (this.max ~= max) then
		this.max = max;
		ConverseFrameHistoryScrollFrameScrollBar:SetValue(max);
	end
end

function ConverseHistory_Update()
	if not RealmName or not PlayerName or RealmName == "" then return; end
	-- Update text
	ConverseFrameHistoryScrollFrameText:SetText(cText);
	if (cNumItems > 0) and (cSelected > 0) then
		ConverseFrameHistoryScrollFrameText:SetTextColor(cItems[cSelected].colour.r, cItems[cSelected].colour.g, cItems[cSelected].colour.b);
	end

	ConverseHistory_UpdateScrollbar();
	if (ConverseState[RealmName][PlayerName].maxlines) and (ConverseState[RealmName][PlayerName].maxlines > 0) then
		ConverseFrameLinesText:SetText(cLines.."/"..ConverseState[RealmName][PlayerName].maxlines);
	else
		ConverseFrameLinesText:SetText(cLines..CONVERSE_HEADER_LINES);
	end

	-- Update buttons
	--
	--if ConverseState[RealmName][PlayerName].historylock then
	--	ConverseFrameLockButtonText:SetText(CONVERSE_BUTTON_UNLOCKHISTORY);
	--	ConverseFrameSaveButton:Disable();
	--else
	--	ConverseFrameLockButtonText:SetText(CONVERSE_BUTTON_LOCKHISTORY);
	--	ConverseFrameSaveButton:Enable();
	--end
end

--function ConverseHistory_SaveEdit()
--	ConverseFrameHistoryScrollFrameText:ClearFocus();
--end

function ConverseHistory_UpdateFromBuffers()
	for key, value in pairs(ConverseHistory[RealmName]) do
		ConverseHistory_UpdateItemFromBuffers(key);
	end
end

function ConverseHistory_UpdateItemFromBuffers(key)
	local i;

	if not ConverseHistory[RealmName][key].buffer then
		ConverseHistory[RealmName][key].buffer = {}
		return;
	end

	for i = 1, table.getn(ConverseHistory[RealmName][key].buffer) do
		-- Check if we've hit the line maximum
		local numlines = table.getn(ConverseHistory[RealmName][key].text);
		if (ConverseState[RealmName][PlayerName].maxlines > 0) and 
		   (numlines > ConverseState[RealmName][PlayerName].maxlines) then
			table.remove(ConverseHistory[RealmName][key].text, numlines - ConverseState[RealmName][PlayerName].maxlines + 1);
		end

		table.insert(ConverseHistory[RealmName][key].text, ConverseHistory[RealmName][key].buffer[i]);
	end
	ConverseHistory[RealmName][key].buffer = {};
end

--=== Ignores ==============================================================--

function ConverseIgnore_OnClick()
	local checked	= this:GetChecked();
	local id	= this:GetID();
	local key;

	if (id == CONVERSE_IGNORE_ID_WHISPER) then
		key = CONVERSE_IGNORE_WHISPER;
	elseif (id == CONVERSE_IGNORE_ID_SAY) then
		key = CONVERSE_IGNORE_SAY;
	elseif (id == CONVERSE_IGNORE_ID_PARTY) then
		key = CONVERSE_IGNORE_PARTY;
	elseif (id == CONVERSE_IGNORE_ID_RAID) then
		key = CONVERSE_IGNORE_RAID;
	elseif (id == CONVERSE_IGNORE_ID_BG) then
		key = CONVERSE_IGNORE_BG;
	elseif (id == CONVERSE_IGNORE_ID_GUILD) then
		key = CONVERSE_IGNORE_GUILD;
	elseif (id == CONVERSE_IGNORE_ID_OFFICER) then
		key = CONVERSE_IGNORE_OFFICER;
	else
		key = id;
	end

	if (checked) then
		ConverseState[RealmName][PlayerName].ignore[key] = 1;
		PlaySound("igMainMenuOptionCheckBoxOff");
	else
		ConverseState[RealmName][PlayerName].ignore[key] = nil;
		PlaySound("igMainMenuOptionCheckBoxOn");
	end

	-- Refresh

	-- FIXME: THIS IS CURRENTLY NOT WORKING!

	local cnum,cname = GetChannelName(id);
	if (cnum > 0 and id == cnum) then
		cSelected = 1;
		ConverseState[RealmName][PlayerName].lastselected = CONVERSE_DEF_SELECTED;
	end	

	ConverseItems_Rebuild();
	ConverseItems_Update();

	ConverseHistory_GenerateText(ConverseState[RealmName][PlayerName].lastselected);
	ConverseHistory_Update();
end

--=== ChatFrameEditBox hook ================================================--

function Converse_ChatFrame_OpenChat(text, chatframe)
	if not ConverseState[RealmName][PlayerName].syncchat then
		return;
	end

	if not chatframe then
		chatframe = DEFAULT_CHAT_FRAME;
	end

	if (cNumItems <= 0) then return; end
	if not cItems[cSelected] or not cItems[cSelected].key then return; end

	local ebox	= chatframe.editBox;
	local prefix 	= string.sub(cItems[cSelected].key, 1, 2);
	local name	= string.sub(cItems[cSelected].key, 3);

	--ConverseItems_Rebuild();
	if not ConverseFrame:IsVisible() then
		return;
	end

	-- Do this only if the main frame is shown
	if (cItems[cSelected].key == CONVERSE_INDEX_SAY) then
		ebox:SetAttribute("chatType", "SAY");
	elseif (cItems[cSelected].key == CONVERSE_INDEX_PARTY) then
		ebox:SetAttribute("chatType", "PARTY");
	elseif (cItems[cSelected].key == CONVERSE_INDEX_RAID) then
		ebox:SetAttribute("chatType", "RAID");
	elseif (cItems[cSelected].key == CONVERSE_INDEX_BG) then
		ebox:SetAttribute("chatType", "BATTLEGROUND");
	elseif (prefix == CONVERSE_PREFIX_GUILD) then
		ebox:SetAttribute("chatType", "GUILD");
	elseif (prefix== CONVERSE_PREFIX_OFFICER) then
		ebox:SetAttribute("chatType", "OFFICER");
	elseif (prefix == CONVERSE_PREFIX_CHANNEL) and cChannels[name] then
		ebox:SetAttribute("chatType", "CHANNEL");
		ebox:SetAttribute("channelTarget", cChannels[name]);
	elseif (prefix == CONVERSE_PREFIX_MSG) then
		ebox:SetAttribute("chatType", "WHISPER");
		ebox:SetAttribute("tellTarget", name);
	elseif (prefix == CONVERSE_PREFIX_MSGRID) then
		ebox:SetAttribute("chatType", "BN_WHISPER");
		ebox:SetAttribute("tellTarget", name);
	else
		-- Fallback
		ebox:SetAttribute("chatType", "SAY");
	end
	ChatEdit_UpdateHeader(ebox);
end

--=== Links ================================================================--

function ConverseLinks_Toggle()
	PlaySound("igMainMenuOptionCheckBoxOn");
	if ConverseFrameLinksPanel:IsVisible() then
		ConverseFrameLinksPanel:Hide();
		ConverseState[RealmName][PlayerName].links = nil;
	else
		ConverseFrameLinksPanel:Show();
		ConverseState[RealmName][PlayerName].links = 1;
	end
end

function ConverseLinks_Hide()
	ConverseFrameLinksPanel:Hide();
	ConverseState[RealmName][PlayerName].links = nil;
end

function ConverseLinks_OnShow()
	ConverseLinks_Update();
end

--function ConverseLinks_OnHide()
--	ConverseState[RealmName][PlayerName].links = nil;
--end

function ConverseLinks_Update()
	if not ConverseFrameLinksPanel:IsVisible() then
		return;
	end

	local i;
	local item;
	local links = {};
	local cname;

	-- Rebuild links list
	if (cNumItems > 0) and (cSelected > 0) then
		local pfx = string.sub(cItems[cSelected].key, 1, 2);
		local cname = string.sub(cItems[cSelected].key, 3);
		if (pfx == CONVERSE_PREFIX_CHANNEL) and cChannelNames[cname] then
			-- List of players
			local pname, tag;
			for pname, tag in pairs(cChannelNames[cname]) do
				if (tag == CONVERSE_TAG_NORMAL) then
					table.insert(links, string.format(CONVERSE_FORMAT_NAME, pname, pname));
				else
					table.insert(links, string.format(CONVERSE_FORMAT_NAME, pname, pname).." "..CONVERSE_COLOUR_TAG..tag);
				end
			end
			-- Player name list will be sorted
			table.sort(links);
		else
			for i=1,table.getn(ConverseHistory[RealmName][cItems[cSelected].key].text) do
				-- Format: |cffa335ee|Hitem:21506:0:0:0|h[Pendant of the Shifting Sands]|h|r
				for item in string.gmatch(ConverseHistory[RealmName][cItems[cSelected].key].text[i], "|c%x+|H[^\]]*]|h|r") do
					-- Don't add duplicates
					if (item ~= links[table.getn(links)]) then
						table.insert(links, item);
					end
				end
			end

		end
	end

	-- Clear the window, and print
	ConverseFrameLinksPanelMessageFrame:Clear();
	for i=1,table.getn(links) do
		ConverseFrameLinksPanelMessageFrame:AddMessage(links[i]);
	end

	if (table.getn(links) > 0) then
		-- Too bad the links end up at the bottom... so let's scroll them up.
		if ConverseFrameLinksPanelMessageFrame:AtTop() then
			while ConverseFrameLinksPanelMessageFrame:AtTop() do
				ConverseFrameLinksPanelMessageFrame:AddMessage(" ");
			end
			-- Okay, we've gone one line over...
			ConverseLinks_ScrollUp();
			ConverseFrameLinksPanelUpButton:Disable();
			ConverseFrameLinksPanelDownButton:Disable();
		else
			-- Long history and we're at the bottom
			ConverseFrameLinksPanelUpButton:Enable();
			ConverseFrameLinksPanelDownButton:Disable();
		end
	else
		-- No links
		ConverseFrameLinksPanelUpButton:Disable();
		ConverseFrameLinksPanelDownButton:Disable();
	end
end

function ConverseLinks_ScrollDown()
	ConverseFrameLinksPanelMessageFrame:ScrollDown();
	if ConverseFrameLinksPanelMessageFrame:AtBottom() then
		ConverseFrameLinksPanelDownButton:Disable();
	end
	ConverseFrameLinksPanelUpButton:Enable();
end

function ConverseLinks_ScrollUp()
	ConverseFrameLinksPanelMessageFrame:ScrollUp();
	if ConverseFrameLinksPanelMessageFrame:AtTop() then
		ConverseFrameLinksPanelUpButton:Disable();
	end
	ConverseFrameLinksPanelDownButton:Enable();
end

function ConverseLinks_OnUpdate(elapsed)
	if (this:GetButtonState() == "PUSHED") then
		-- Bah, who cares about clickdelay :P
		local name = this:GetName();
		if (name == "ConverseFrameLinksPanelUpButton") then
			ConverseLinks_ScrollUp();
		elseif (name == "ConverseFrameLinksPanelDownButton") then
			ConverseLinks_ScrollDown();
		end
	end
end

--=== Window Handling ======================================================--

function Converse_ToggleVisibility()
	if (ConverseFrame:IsVisible()) then
		ConverseFrame:Hide();
		ConverseState[RealmName][PlayerName].hidden = 1;
		PlaySound("igQuestListClose");
	else
		ConverseFrame:Show();
		ConverseState[RealmName][PlayerName].hidden = nil;
		PlaySound("igQuestListOpen");
	end
end

function Converse_OnDragStart()
	if not ConverseState[RealmName][PlayerName].locked then
		ConverseFrame:StartMoving();
	end
end

function Converse_OnDragStop()
	ConverseFrame:StopMovingOrSizing();
end

function ConverseLock_Toggle()
	local checked = this:GetChecked();
	if (checked) then
		ConverseState[RealmName][PlayerName].locked = 1;
		PlaySound("igMainMenuOptionCheckBoxOff");
	else
		ConverseState[RealmName][PlayerName].locked = nil;
		PlaySound("igMainMenuOptionCheckBoxOn");
	end
end

--=== Alt Edit ==============================================================--

function Converse_GetCtype()
	if (cNumItems <= 0) then return; end
	if not cItems[cSelected] or not cItems[cSelected].key then return; end

	local prefix 	= string.sub(cItems[cSelected].key, 1, 2);
	local name	= string.sub(cItems[cSelected].key, 3);
	local ctype = nil;
	local ctarget = nil;

	-- Do this only if the main frame is shown
	if (cItems[cSelected].key == CONVERSE_INDEX_SAY) then
		ctype = "SAY";
	elseif (cItems[cSelected].key == CONVERSE_INDEX_PARTY) then
		ctype = "PARTY";
	elseif (cItems[cSelected].key == CONVERSE_INDEX_RAID) then
		ctype = "RAID";
	elseif (cItems[cSelected].key == CONVERSE_INDEX_BG) then
		ctype = "BATTLEGROUND";
	elseif (prefix == CONVERSE_PREFIX_GUILD) then
		ctype = "GUILD";
	elseif (prefix== CONVERSE_PREFIX_OFFICER) then
		ctype = "OFFICER";
	elseif (prefix == CONVERSE_PREFIX_CHANNEL) and cChannels[name] then
		ctype = "CHANNEL";
		--ebox.channelTarget = name;
		ctarget = cChannels[name];
	elseif (prefix == CONVERSE_PREFIX_MSG) then
		ctype = "WHISPER";
		ctarget = name;
	else
		-- Fallback
		ctype = "SAY";
	end

	return ctype, ctarget;
end

function ConverseAltEdit_OnEnterPressed()
	local text = ConverseFrameAltEditFrameText:GetText();
	local ctype, ctarget = Converse_GetCtype();

	-- Send it off
	--cMessage("text = "..text.." // ctype = "..ctype);
	SendChatMessage(text, ctype, nil, ctarget);

	-- Reset the box
	ConverseFrameAltEditFrameText:SetText("");
	ConverseFrameAltEditFrameText:SetFocus();
end

--=== Multi-line Paste ======================================================--

function ConverseMultipaste_Toggle()
	PlaySound("igMainMenuOptionCheckBoxOn");
	if ConverseFrameMultipasteFrame:IsVisible() then
		ConverseFrameMultipasteFrame:Hide();
		ConverseFrameMultipasteFrameScrollFrameText:ClearFocus();
	else
		ConverseFrameMultipasteFrame:Show();
		ConverseFrameMultipasteFrameScrollFrameText:SetFocus();
	end
end

function ConverseMultipaste_OnTextChanged()
	if cFinishedLoading then
		ConverseState[RealmName][PlayerName].multipaste = ConverseFrameMultipasteFrameScrollFrameText:GetText();
	end
end

function ConverseMultipaste_Clear()
	ConverseFrameMultipasteFrameScrollFrameText:SetText("");
	ConverseFrameMultipasteFrameScrollFrameText:SetFocus();
end

function ConverseMultipaste_Paste()
	if cPasting then
		-- We are currently in a paste, so we will wait until that one finishes
		return;
	end

	local text = ConverseFrameMultipasteFrameScrollFrameText:GetText();
	local ctype, ctarget = Converse_GetCtype();
	local oneline = "";

	cPasteCType = ctype;
	cPasteCTarget = ctarget;

	-- Split text into separate lines

	for oneline in string.gmatch(text,"[^\r\n]+") do
		--ChatThrottleLib:SendChatMessage("NORMAL", nil, oneline, ctype, nil, ctarget)
		local linelen = string.len(oneline)
		if linelen > CONVERSE_SPLITLENGTH then
			-- Sorry, we don't do word wrapping at this time!
			local offset = 1;
			while offset < string.len(oneline) do
				local splitline = string.sub(oneline, offset, offset + CONVERSE_SPLITLENGTH - 1);
				ChatThrottleLib:SendChatMessage("NORMAL", "conv", splitline, ctype, nil, ctarget)
				offset = offset + CONVERSE_SPLITLENGTH;
			end
		else
			ChatThrottleLib:SendChatMessage("NORMAL", "conv", oneline, ctype, nil, ctarget)
		end
	end

	-- Done!
	ConverseFrameMultipasteFrame:Hide();
	ConverseFrameMultipasteFrameScrollFrameText:ClearFocus();
end
